import itertools
from typing import List

from music21.bar import Repeat
from music21.chord import Chord
from music21.meter import TimeSignature
from music21.note import Note, Rest
from music21.repeat import RepeatMark
from music21.spanner import RepeatBracket, Slur
from music21.stream import Measure, Part, Score

from musif.logs import ldebug

# TODO: document this module


def measure_ranges(
    instrument_measures,
    init,
    end,
    iteration=None,
    offset=None,
    twoCompasses=False,
    remove_repetition_marks=False,
):
    measures = []
    o = offset
    last_offset = (
        0.0 if int(init) - 6 < 0 else instrument_measures[int(init) - 6].offset
    )

    # Find index where measureNumber init and end is stored
    init_index = instrument_measures.index(
        [m for m in instrument_measures if m.measureNumber == init][0]
    )
    end_compass = [m for m in instrument_measures if m.measureNumber == end]
    end_index = (
        instrument_measures.index(end_compass[0])
        if len(end_compass) > 0
        else len(instrument_measures) - 1
    )
    for i in range(init_index, end_index + 1):
        if (
            not i < 0
            and i < len(instrument_measures)
            and instrument_measures[i].measureNumber >= int(init)
            and instrument_measures[i].measureNumber <= int(end)
        ):
            if not twoCompasses:
                compass = instrument_measures[i].quarterLength
                m = Measure(number=instrument_measures[i].measureNumber)
                if remove_repetition_marks:
                    m.elements = [
                        e
                        for e in instrument_measures[i].elements
                        if not isinstance(e, RepeatMark)
                    ]
                else:
                    m.elements = instrument_measures[i].elements
                m.quarterLength = compass
                if offset is None:
                    m.offset = last_offset
                else:
                    m.offset = o
                    o += compass

                measures.append(m)
                if (
                    instrument_measures[i].measureNumber != 0.0
                    and instrument_measures[i].offset != 0
                ):
                    last_offset = instrument_measures[i].offset + compass
            twoCompasses = False

    """if iteration is 1:
        measures = measures[:-1]
    elif iteration is 2:
        measures = measures[:-2] + [measures[-1]]"""
    if iteration == 2:
        last_measure = instrument_measures[i + 1]
        last_measure.offset = measures[-1].offset
        measures = measures[:-1] + [last_measure]
    return measures


# TODO: This function is too long
def expand_repeat_bars(score):
    final_score = Score()
    final_score.metadata = score.metadata
    exist_repetition_bars = False
    # find repeat bars and expand
    for instr in score.parts:
        part_measures = get_measures_with_repetitions(
            instr.elements
        )  # returns the measures with repetitions
        last_measure = part_measures[-1].measureNumber
        part_measures_expanded = []
        startsin0 = part_measures[0].measureNumber == 0  # Everything should be -1
        repetition_bars = []

        # Find all repetitions in that instrument
        for elem in instr.elements:
            if isinstance(elem, Measure):
                for e in elem:
                    if isinstance(e, Repeat):
                        exist_repetition_bars = True
                        if e.direction == "start":
                            repetition_bars.append((e.measureNumber, "start"))
                        elif e.direction == "end":
                            repetition_bars.append((e.measureNumber, "end"))
                        index = elem.elements.index(e)
                        elem.elements = (
                            elem.elements[:index] + elem.elements[index + 1 :]
                        )
            elif isinstance(elem, RepeatBracket):
                string_e = str(elem)
                index = string_e.find("music21.stream.Measure")
                measure = (
                    string_e[index:].replace("music21.stream.Measure", "")[1:3].strip()
                )
                repetition_bars.append((int(measure), elem.number))
                index = instr.elements.index(elem)
                elem.elements = instr.elements[:index] + instr.elements[index + 1 :]
        repetition_bars = sorted(list(repetition_bars), key=lambda tup: tup[0])

        start = 0 if startsin0 else 1
        if exist_repetition_bars:
            p = Part()
            p.id = instr.id
            p.partName = instr.partName
            for rb in repetition_bars:
                compass = measure_ranges(part_measures, rb[0], rb[0])[0].quarterLength
                if rb[1] == "start":
                    if len(part_measures_expanded) > 0:
                        offset = part_measures_expanded[-1][-1].offset
                    else:
                        offset = 0
                    start_measures = measure_ranges(
                        part_measures, start, rb[0] - 1, offset=offset + compass
                    )
                    if len(start_measures) > 0:
                        part_measures_expanded.append(start_measures)
                    start = rb[0]
                elif rb[1] == "end":
                    if len(part_measures_expanded) > 0:
                        offset = part_measures_expanded[-1][-1].offset
                    else:
                        offset = 0
                    casilla_1 = (
                        True
                        if any(
                            re[1] == "1" and re[0] <= rb[0] for re in repetition_bars
                        )
                        else False
                    )
                    casilla_2 = None
                    if casilla_1:
                        casilla_2 = [
                            re
                            for re in repetition_bars
                            if re[1] == "2" and re[0] > rb[0]
                        ]
                        casilla_2 = None if len(casilla_2) == 0 else casilla_2[0]
                    part_measures_expanded.append(
                        measure_ranges(
                            part_measures,
                            init=start,
                            end=rb[0],
                            offset=offset + compass,
                            remove_repetition_marks=True,
                        )
                    )  # This should erase the repetition marks
                    if casilla_2 is not None:
                        part_measures_expanded.append(
                            measure_ranges(
                                part_measures,
                                start,
                                casilla_2[0],
                                iteration=2,
                                offset=part_measures_expanded[-1][-1].offset + compass,
                            )
                        )
                        start = casilla_2[0] + 1
                    else:
                        part_measures_expanded.append(
                            measure_ranges(
                                part_measures,
                                init=start,
                                end=rb[0],
                                offset=part_measures_expanded[-1][-1].offset + compass,
                            )
                        )
                        start = rb[0] + 1
            if start < last_measure:
                compass = measure_ranges(part_measures, start, start + 1)[
                    0
                ].quarterLength
                offset = part_measures_expanded[-1][-1].offset
                part_measures_expanded.append(
                    measure_ranges(
                        part_measures, start, last_measure + 1, offset=offset + compass
                    )
                )
            p.elements = list(itertools.chain(*tuple(part_measures_expanded)))
            final_score.insert(0, p)

    return final_score if exist_repetition_bars else score


def slur_processing(part):
    slurs = [s for s in part.elements if isinstance(s, Slur)]
    part_measures = get_measures_with_repetitions(
        part.elements
    )  # returns the measures with repetitions

    for slur in slurs:
        slur_notes = slur.getSpannedElements()
        first_note_measure = slur_notes[0].measureNumber
        first_note_offset = slur_notes[0].offset
        last_note_measure = slur_notes[-1].measureNumber
        last_note_offset = slur_notes[-1].offset
        # Buscar ese compas y offset y recorrer desde esa nota hasta la siguiente
        implied_measures = measure_ranges(
            part_measures, first_note_measure, last_note_measure
        )
        first_found = False
        last_found = False
        for measure in implied_measures:
            notes = measure.elements
            i = 0
            while i < len(notes) and not last_found:
                if isinstance(notes[i], Note):
                    if notes[i].offset == first_note_offset:
                        first_found = True
                    elif notes[i].offset == last_note_offset:
                        last_found = True
                    elif first_found and not last_found:
                        slur.addSpannedElements(notes[i])
                i += 1


def expand_score_repetitions(score, repeat_elements):
    score = expand_repeat_bars(score)  # FIRST EXPAND REPEAT BARS
    final_score = Score()
    final_score.metadata = score.metadata

    if len(repeat_elements) > 0:
        for part in score.parts:
            p = expand_part(part, repeat_elements)

            final_score.insert(0, p)
    else:
        final_score = score
    return final_score


# TODO: this function seems alittle long as well
def get_expanded_measures(part_measures, repeat_elements):
    repeat_bracket = (
        sum([1 if "repeat bracket" in item[1] else 0 for item in repeat_elements]) > 0
    )
    measures_list = []
    startsin0 = part_measures[0].measureNumber == 0  # Everything should be -1

    there_is_fine = False
    there_is_segno = False
    # 1. find the fine and segno
    if any([r[1] == "fine" for r in repeat_elements]):
        f = [x[0] for x in repeat_elements if x[1] == "fine"][0]
        there_is_fine = True
    if any([r[1] == "segno" for r in repeat_elements]):
        s = [x[0] for x in repeat_elements if x[1] == "segno"][0]
        there_is_segno = True

    # 2. Having all the repetition elements, get the measures
    if there_is_segno:  # Introduction
        before_segno = measure_ranges(part_measures, 1 if not startsin0 else 0, s - 1)
        measures_list.append(
            before_segno
        )  # S -1 OR S-> when segno in compass 1, s, else s-1?
        dc_time_signature = [
            y for x in before_segno for y in x.elements if isinstance(y, TimeSignature)
        ]
    elif there_is_fine:
        measures_list.append(
            measure_ranges(
                part_measures,
                1 if not startsin0 else 0,
                f - 1,
                iteration=1 if repeat_bracket else None,
            )
        )
    else:
        measures_list.append(
            measure_ranges(
                part_measures,
                1 if not startsin0 else 0,
                len(part_measures),
                iteration=1 if repeat_bracket else None,
            )
        )

    for repeat in repeat_elements:
        repeat_measure = measure_ranges(part_measures, repeat[0], repeat[0])
        compass = repeat_measure[0].quarterLength

        if repeat[1] == "segno":
            offset = measures_list[-1][-1].offset
            segno_part = measure_ranges(
                part_measures,
                s,
                f - 1 if there_is_fine else len(part_measures),
                iteration=1 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks=True,
            )
            measures_list.append(segno_part)  # Segno to Fine
        elif repeat[1] == "fine":
            twoCompasses = False
            """if len(repeat_measure) > 0:
                twoCompasses = True"""
            offset = measures_list[-1][-1].offset
            fine_part = measure_ranges(
                part_measures,
                f,
                len(part_measures),
                offset=offset + compass,
                twoCompasses=twoCompasses,
                remove_repetition_marks=True,
            )
            measures_list.append(fine_part)  # Fine to end
        elif repeat[1] == "al segno" or repeat[1] == "dal segno":
            offset = measures_list[-1][-1].offset
            # segnos' compass time signature
            segno_time_measure = [
                x for x in segno_part[0].elements if isinstance(x, TimeSignature)
            ]
            segno_time_measure = (
                segno_time_measure
                if len(segno_time_measure) != 0
                else dc_time_signature[-1]
            )
            alsegno_list = measure_ranges(
                part_measures,
                s,
                f - 1 if there_is_fine else len(part_measures),
                iteration=2 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks=True,
            )
            if not any(isinstance(x, TimeSignature) for x in alsegno_list[0].elements):
                # we reset the time signature that was on the dacapo
                alsegno_list[0].elements = tuple(
                    [segno_time_measure] + list(alsegno_list[0].elements)
                )

            measures_list.append(alsegno_list)  # Segno to fine
        elif repeat[1] == "da capo":
            offset = measures_list[-1][-1].offset
            if startsin0 and there_is_fine and not repeat_bracket:
                f += 1
            dacapo_list = measure_ranges(
                part_measures,
                0 if startsin0 else 1,
                f - 1 if there_is_fine else len(part_measures),
                iteration=2 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks=True,
            )
            measures_list.append(dacapo_list)

    return measures_list


def get_measures_with_repetitions(obj) -> List[Measure]:
    measures = []
    for elem in obj:
        if isinstance(elem, Measure):
            measures.append(elem)
            # Change the note offsets to avoid problems due to slurs
            last_offset = 0
            last_duration = 0
            for note in elem:
                if (
                    isinstance(note, Note)
                    or isinstance(note, Rest)
                    or isinstance(note, Chord)
                ):
                    note.offset = last_offset + last_duration
                    last_offset = note.offset
                    last_duration = note.duration.quarterLength
    return measures


def expand_part(part: Part, repeat_elements):
    part_measures = get_measures_with_repetitions(
        part.elements
    )  # returns measures with repetitions
    p = Part()
    p.id = part.id
    p.partName = part.partName

    part_measures_expanded = get_expanded_measures(
        part_measures, repeat_elements
    )  # returns the measures expanded
    part_measures_expanded = list(itertools.chain(*part_measures_expanded))
    # part_measures_expanded = sorted(tuple(part_measures_expanded), key =lambda x: x.offset)
    # Assign a new continuous compass number to every measure
    measure_number = 0 if part_measures_expanded[0].measureNumber == 0 else 1
    for i, e in enumerate(part_measures_expanded):
        m = Measure(number=measure_number)
        m.elements = e.elements
        m.offset = e.offset
        m.quarterLength = e.quarterLength
        part_measures_expanded[i] = m
        measure_number += 1
    p.elements = part_measures_expanded
    return p


def get_repetition_elements(score: Score, verbose=True) -> List[tuple]:
    # 1. Get the repeat elements
    repeat_elements = set()

    for part in score.parts:
        instr_repeat_elements = []
        for elem in part.elements:
            if isinstance(elem, Measure):
                for e in elem:
                    if isinstance(e, RepeatMark) and not isinstance(e, Repeat):
                        measure = e.measureNumber
                        if elem.numberSuffix in ["X1", "X2"]:  # Exception
                            measure += 1
                        instr_repeat_elements.append((measure, e.name))
            elif isinstance(elem, RepeatBracket):
                string_e = str(elem)
                index = string_e.find("music21.stream.Measure")
                string_e = string_e[index:].replace("music21.stream.Measure ", "")
                measure = (
                    string_e.split(" ")[0].strip().replace("X1", "").replace("X2", "")
                )
                instr_repeat_elements.append(
                    (int(measure), "repeat bracket" + elem.number)
                )
        repeat_elements.update(instr_repeat_elements)

    repeat_elements = sorted(list(repeat_elements), key=lambda tup: tup[0])
    if verbose:
        ldebug(f"The repeat elements found in this score are: {str(repeat_elements)}")
    return repeat_elements



MEAN_INTERVAL = "MeanInterval"
INTERVALLIC_MEAN = "IntervallicMean"
INTERVALLIC_STD = "IntervallicStd"
ABSOLUTE_INTERVALLIC_MEAN = "AbsoluteIntervallicMean"
ABSOLUTE_INTERVALLIC_STD = "AbsoluteIntervallicStd"
ASCENDING_INTERVALLIC_MEAN = "AscendingIntervallicMean"
ASCENDING_INTERVALLIC_STD = "AscendingIntervallicStd"
DESCENDING_INTERVALLIC_MEAN = "DescendingIntervallicMean"
DESCENDING_INTERVALLIC_STD = "DescendingIntervallicStd"
TRIMMED_INTERVALLIC_MEAN = "TrimmedIntervallicMean"
TRIMMED_INTERVALLIC_STD = "TrimmedIntervallicStd"
TRIMMED_ABSOLUTE_INTERVALLIC_MEAN = "TrimmedAbsoluteIntervallicMean"
TRIMMED_ABSOLUTE_INTERVALLIC_STD = "TrimmedAbsoluteIntervallicStd"
INTERVALLIC_TRIM_DIFF = "IntervallicTrimDiff"
INTERVALLIC_TRIM_RATIO = "IntervallicTrimRatio"
ABSOLUTE_INTERVALLIC_TRIM_DIFF = "AbsoluteIntervallicTrimDiff"
ABSOLUTE_INTERVALLIC_TRIM_RATIO = "AbsoluteIntervallicTrimRatio"
ASCENDING_SEMITONES_SUM = "AscendingSemitones_Sum"
DESCENDING_SEMITONES_SUM = "DescendingSemitones_Sum"
ASCENDING_INTERVALS_COUNT = "AscendingIntervals_Count"
ASCENDING_INTERVALS_PER = "AscendingIntervals_Per"
DESCENDING_INTERVALS_COUNT = "DescendingIntervals_Count"
DESCENDING_INTERVALS_PER = "DescendingIntervals_Per"
REPEATED_NOTES_COUNT = "RepeatedNotes_Count"
REPEATED_NOTES_PER = "RepeatedNotes_Per"
LEAPS_ASC_COUNT = "LeapsAsc_Count"
LEAPS_ASC_PER = "LeapsAsc_Per"
LEAPS_DESC_COUNT = "LeapsDesc_Count"
LEAPS_DESC_PER = "LeapsDesc_Per"
LEAPS_ALL_COUNT = "LeapsAll_Count"
LEAPS_ALL_PER = "LeapsAll_Per"
STEPWISE_MOTION_ASC_COUNT = "StepwiseMotionAsc_Count"
STEPWISE_MOTION_DESC_COUNT = "StepwiseMotionDesc_Count"
STEPWISE_MOTION_ALL_COUNT = "StepwiseMotionAll_Count"
STEPWISE_MOTION_ASC_PER = "StepwiseMotionAsc_Per"
STEPWISE_MOTION_DESC_PER = "StepwiseMotionDesc_Per"
STEPWISE_MOTION_ALL_PER = "StepwiseMotionAll_Per"
INTERVALS = "Intervals"
INTERVALS_PERFECT_ASC_COUNT = INTERVALS + "PerfectAsc_Count"
INTERVALS_PERFECT_DESC_COUNT = INTERVALS + "PerfectDesc_Count"
INTERVALS_PERFECT_ALL_COUNT = INTERVALS + "PerfectAll_Count"
INTERVALS_PERFECT_ASC_PER = INTERVALS + "PerfectAsc_Per"
INTERVALS_PERFECT_DESC_PER = INTERVALS + "PerfectDesc_Per"
INTERVALS_PERFECT_ALL_PER = INTERVALS + "PerfectAll_Per"
INTERVALS_MAJOR_ASC_COUNT = INTERVALS + "MajorAsc_Count"
INTERVALS_MAJOR_DESC_COUNT = INTERVALS + "MajorDesc_Count"
INTERVALS_MAJOR_ALL_COUNT = INTERVALS + "MajorAll_Count"
INTERVALS_MAJOR_ASC_PER = INTERVALS + "MajorAsc_Per"
INTERVALS_MAJOR_DESC_PER = INTERVALS + "MajorDesc_Per"
INTERVALS_MAJOR_ALL_PER = INTERVALS + "MajorAll_Per"
INTERVALS_MINOR_ASC_COUNT = INTERVALS + "MinorAsc_Count"
INTERVALS_MINOR_DESC_COUNT = INTERVALS + "MinorDesc_Count"
INTERVALS_MINOR_ALL_COUNT = INTERVALS + "MinorAll_Count"
INTERVALS_MINOR_ASC_PER = INTERVALS + "MinorAsc_Per"
INTERVALS_MINOR_DESC_PER = INTERVALS + "MinorDesc_Per"
INTERVALS_MINOR_ALL_PER = INTERVALS + "MinorAll_Per"
INTERVALS_AUGMENTED_ASC_COUNT = INTERVALS + "AugmentedAsc_Count"
INTERVALS_AUGMENTED_DESC_COUNT = INTERVALS + "AugmentedDesc_Count"
INTERVALS_AUGMENTED_ALL_COUNT = INTERVALS + "AugmentedAll_Count"
INTERVALS_AUGMENTED_ASC_PER = INTERVALS + "AugmentedAsc_Per"
INTERVALS_AUGMENTED_DESC_PER = INTERVALS + "AugmentedDesc_Per"
INTERVALS_AUGMENTED_ALL_PER = INTERVALS + "AugmentedAll_Per"
INTERVALS_DIMINISHED_ASC_COUNT = INTERVALS + "DiminishedAsc_Count"
INTERVALS_DIMINISHED_DESC_COUNT = INTERVALS + "DiminishedDesc_Count"
INTERVALS_DIMINISHED_ALL_COUNT = INTERVALS + "DiminishedAll_Count"
INTERVALS_DIMINISHED_ASC_PER = INTERVALS + "DiminishedAsc_Per"
INTERVALS_DIMINISHED_DESC_PER = INTERVALS + "DiminishedDesc_Per"
INTERVALS_DIMINISHED_ALL_PER = INTERVALS + "DiminishedAll_Per"
INTERVALS_DOUBLE_AUGMENTED_ASC_COUNT = INTERVALS + "DoubleAugmentedAsc_Count"
INTERVALS_DOUBLE_AUGMENTED_DESC_COUNT = INTERVALS + "DoubleAugmentedDesc_Count"
INTERVALS_DOUBLE_AUGMENTED_ALL_COUNT = INTERVALS + "DoubleAugmentedAll_Count"
INTERVALS_DOUBLE_AUGMENTED_ASC_PER = INTERVALS + "DoubleAugmentedAsc_Per"
INTERVALS_DOUBLE_AUGMENTED_DESC_PER = INTERVALS + "DoubleAugmentedDesc_Per"
INTERVALS_DOUBLE_AUGMENTED_ALL_PER = INTERVALS + "DoubleAugmentedAll_Per"
INTERVALS_DOUBLE_DIMINISHED_ASC_COUNT = INTERVALS + "DoubleDiminishedAsc_Count"
INTERVALS_DOUBLE_DIMINISHED_DESC_COUNT = INTERVALS + "DoubleDiminishedDesc_Count"
INTERVALS_DOUBLE_DIMINISHED_ALL_COUNT = INTERVALS + "DoubleDiminishedAll_Count"
INTERVALS_DOUBLE_DIMINISHED_ASC_PER = INTERVALS + "DoubleDiminishedAsc_Per"
INTERVALS_DOUBLE_DIMINISHED_DESC_PER = INTERVALS + "DoubleDiminishedDesc_Per"
INTERVALS_DOUBLE_DIMINISHED_ALL_PER = INTERVALS + "DoubleDiminishedAll_Per"
INTERVALS_WITHIN_OCTAVE_ASC_COUNT = INTERVALS + "WithinOctaveAsc_Count"
INTERVALS_WITHIN_OCTAVE_DESC_COUNT = INTERVALS + "WithinOctaveDesc_Count"
INTERVALS_WITHIN_OCTAVE_ALL_COUNT = INTERVALS + "WithinOctaveAll_Count"
INTERVALS_WITHIN_OCTAVE_ASC_PER = INTERVALS + "WithinOctaveAsc_Per"
INTERVALS_WITHIN_OCTAVE_DESC_PER = INTERVALS + "WithinOctaveDesc_Per"
INTERVALS_WITHIN_OCTAVE_ALL_PER = INTERVALS + "WithinOctaveAll_Per"
INTERVALS_BEYOND_OCTAVE_ASC_COUNT = INTERVALS + "BeyondOctaveAsc_Count"
INTERVALS_BEYOND_OCTAVE_DESC_COUNT = INTERVALS + "BeyondOctaveDesc_Count"
INTERVALS_BEYOND_OCTAVE_ALL_COUNT = INTERVALS + "BeyondOctaveAll_Count"
INTERVALS_BEYOND_OCTAVE_ASC_PER = INTERVALS + "BeyondOctaveAsc_Per"
INTERVALS_BEYOND_OCTAVE_DESC_PER = INTERVALS + "BeyondOctaveDesc_Per"
INTERVALS_BEYOND_OCTAVE_ALL_PER = INTERVALS + "BeyondOctaveAll_Per"
INTERVALLIC_SKEWNESS = "IntervallicSkewness"
INTERVALLIC_KURTOSIS = "IntervallicKurtosis"
ABSOLUTE_INTERVALLIC_SKEWNESS = "AbsoluteIntervallicSkewness"
ABSOLUTE_INTERVALLIC_KURTOSIS = "AbsoluteIntervallicKurtosis"
INTERVAL_COUNT = "{prefix}Interval{interval}_Count"
INTERVAL_PER = "{prefix}Interval{interval}_Per"
LARGEST_INTERVAL_ASC = "LargestIntervalAsc"
LARGEST_SEMITONES_ASC = "LargestSemitonesAsc"
LARGEST_ABSOLUTE_SEMITONES_ASC = "LargestAbsoluteSemitonesAsc"
LARGEST_INTERVAL_DESC = "LargestIntervalDesc"
LARGEST_SEMITONES_DESC = "LargestSemitonesDesc"
LARGEST_ABSOLUTE_SEMITONES_DESC = "LargestAbsoluteSemitonesDesc"
LARGEST_INTERVAL_ALL = "LargestIntervalAll"
LARGEST_SEMITONES_ALL = "LargestSemitonesAll"
LARGEST_ABSOLUTE_SEMITONES_ALL = "LargestAbsoluteSemitonesAll"
SMALLEST_INTERVAL_ASC = "SmallestIntervalAsc"
SMALLEST_SEMITONES_ASC = "SmallestSemitonesAsc"
SMALLEST_ABSOLUTE_SEMITONES_ASC = "SmallestAbsoluteSemitonesAsc"
SMALLEST_INTERVAL_DESC = "SmallestIntervalDesc"
SMALLEST_SEMITONES_DESC = "SmallestSemitonesDesc"
SMALLEST_ABSOLUTE_SEMITONES_DESC = "SmallestAbsoluteSemitonesDesc"
SMALLEST_INTERVAL_ALL = "SmallestIntervalAll"
SMALLEST_SEMITONES_ALL = "SmallestSemitonesAll"
SMALLEST_ABSOLUTE_SEMITONES_ALL = "SmallestAbsoluteSemitonesAll"

SPEED_AVG_ABS = "Spe_avg_abs"
ACCELERATION_AVG_ABS = "Acc_avg_abs"
ASCENDENT_AVERAGE = "Asc_avg"
DESCENDENT_AVERAGE = "Dsc_avg"
ASCENDENT_PROPORTION = "Asc_prp"
DESCENDENT_PROPORTION = "Dsc_prp"
MOTION_STEPS = [0.125, 0.25, 0.5, 1.0, 2.0, 4.0, 8.0]
MOTION_WINS = [2, 4, 8, 16, 32]

SCORE_FEATURES = [
    MEAN_INTERVAL,
    INTERVALLIC_MEAN,
    INTERVALLIC_STD,
    ABSOLUTE_INTERVALLIC_MEAN,
    ABSOLUTE_INTERVALLIC_STD,
    ASCENDING_INTERVALLIC_MEAN,
    ASCENDING_INTERVALLIC_STD,
    DESCENDING_INTERVALLIC_MEAN,
    DESCENDING_INTERVALLIC_STD,
    TRIMMED_INTERVALLIC_MEAN,
    TRIMMED_INTERVALLIC_STD,
    TRIMMED_ABSOLUTE_INTERVALLIC_MEAN,
    TRIMMED_ABSOLUTE_INTERVALLIC_STD,
    ABSOLUTE_INTERVALLIC_TRIM_DIFF,
    ABSOLUTE_INTERVALLIC_TRIM_RATIO,
    ASCENDING_INTERVALS_COUNT,
    DESCENDING_INTERVALS_COUNT,
    ASCENDING_INTERVALS_PER,
    DESCENDING_SEMITONES_SUM,
    REPEATED_NOTES_COUNT,
    REPEATED_NOTES_PER,
    LEAPS_ASC_COUNT,
    LEAPS_DESC_COUNT,
    LEAPS_ALL_COUNT,
    LEAPS_ASC_PER,
    LEAPS_DESC_PER,
    LEAPS_ALL_PER,
    LARGEST_INTERVAL_ASC,
    LARGEST_SEMITONES_ASC,
    LARGEST_ABSOLUTE_SEMITONES_ASC,
    LARGEST_INTERVAL_DESC,
    LARGEST_SEMITONES_DESC,
    LARGEST_ABSOLUTE_SEMITONES_DESC,
    LARGEST_INTERVAL_ALL,
    LARGEST_SEMITONES_ALL,
    LARGEST_ABSOLUTE_SEMITONES_ALL,
    SMALLEST_INTERVAL_ASC,
    SMALLEST_SEMITONES_ASC,
    SMALLEST_ABSOLUTE_SEMITONES_ASC,
    SMALLEST_INTERVAL_DESC,
    SMALLEST_SEMITONES_DESC,
    SMALLEST_ABSOLUTE_SEMITONES_DESC,
    SMALLEST_INTERVAL_ALL,
    SMALLEST_SEMITONES_ALL,
    SMALLEST_ABSOLUTE_SEMITONES_ALL,
    STEPWISE_MOTION_ASC_COUNT,
    STEPWISE_MOTION_DESC_COUNT,
    STEPWISE_MOTION_ALL_COUNT,
    STEPWISE_MOTION_ASC_PER,
    STEPWISE_MOTION_DESC_PER,
    STEPWISE_MOTION_ALL_PER,
    INTERVALS_PERFECT_ASC_COUNT,
    INTERVALS_PERFECT_DESC_COUNT,
    INTERVALS_PERFECT_ALL_COUNT,
    INTERVALS_PERFECT_ASC_PER,
    INTERVALS_PERFECT_DESC_PER,
    INTERVALS_PERFECT_ALL_PER,
    INTERVALS_MAJOR_ASC_COUNT,
    INTERVALS_MAJOR_DESC_COUNT,
    INTERVALS_MAJOR_ALL_COUNT,
    INTERVALS_MAJOR_ASC_PER,
    INTERVALS_MAJOR_DESC_PER,
    INTERVALS_MAJOR_ALL_PER,
    INTERVALS_MINOR_ASC_COUNT,
    INTERVALS_MINOR_DESC_COUNT,
    INTERVALS_MINOR_ALL_COUNT,
    INTERVALS_MINOR_ASC_PER,
    INTERVALS_MINOR_DESC_PER,
    INTERVALS_MINOR_ALL_PER,
    INTERVALS_AUGMENTED_ASC_COUNT,
    INTERVALS_AUGMENTED_DESC_COUNT,
    INTERVALS_AUGMENTED_ALL_COUNT,
    INTERVALS_AUGMENTED_ASC_PER,
    INTERVALS_AUGMENTED_DESC_PER,
    INTERVALS_AUGMENTED_ALL_PER,
    INTERVALS_DIMINISHED_ASC_COUNT,
    INTERVALS_DIMINISHED_DESC_COUNT,
    INTERVALS_DIMINISHED_ALL_COUNT,
    INTERVALS_DIMINISHED_ASC_PER,
    INTERVALS_DIMINISHED_DESC_PER,
    INTERVALS_DIMINISHED_ALL_PER,
    INTERVALS_DOUBLE_AUGMENTED_ALL_COUNT,
    INTERVALS_DOUBLE_AUGMENTED_ASC_COUNT,
    INTERVALS_DOUBLE_AUGMENTED_DESC_COUNT,
    INTERVALS_DOUBLE_AUGMENTED_ALL_PER,
    INTERVALS_DOUBLE_AUGMENTED_ASC_PER,
    INTERVALS_DOUBLE_AUGMENTED_DESC_PER,
    INTERVALS_DOUBLE_DIMINISHED_ALL_COUNT,
    INTERVALS_DOUBLE_DIMINISHED_ASC_COUNT,
    INTERVALS_DOUBLE_DIMINISHED_DESC_COUNT,
    INTERVALS_DOUBLE_DIMINISHED_ALL_PER,
    INTERVALS_DOUBLE_DIMINISHED_ASC_PER,
    INTERVALS_DOUBLE_DIMINISHED_DESC_PER,
    INTERVALS_WITHIN_OCTAVE_ALL_COUNT,
    INTERVALS_WITHIN_OCTAVE_ALL_PER,
    INTERVALS_WITHIN_OCTAVE_ASC_COUNT,
    INTERVALS_WITHIN_OCTAVE_ASC_PER,
    INTERVALS_WITHIN_OCTAVE_DESC_COUNT,
    INTERVALS_WITHIN_OCTAVE_DESC_PER,
    INTERVALS_BEYOND_OCTAVE_ALL_COUNT,
    INTERVALS_BEYOND_OCTAVE_ALL_PER,
    INTERVALS_BEYOND_OCTAVE_ASC_COUNT,
    INTERVALS_BEYOND_OCTAVE_ASC_PER,
    INTERVALS_BEYOND_OCTAVE_DESC_COUNT,
    INTERVALS_BEYOND_OCTAVE_DESC_PER,
    INTERVALLIC_SKEWNESS,
    INTERVALLIC_KURTOSIS,
    ABSOLUTE_INTERVALLIC_SKEWNESS,
    ABSOLUTE_INTERVALLIC_KURTOSIS,
]



import csv
import json
from typing import Iterator, List, Union

import pandas as pd
import yaml
from pandas import DataFrame

from musif.common._constants import CSV_DELIMITER, ENCODING
from musif.common.constants import COLOR_SEQ, COLORS, RESET_SEQ


def write_object_to_json_file(
    obj: Union[dict, list], json_file_path: str, indent: int = 4
):
    with open(json_file_path, "w", encoding=ENCODING) as json_file:
        json.dump(obj, json_file, indent=indent)


def read_object_from_json_file(json_file_path: str):
    with open(json_file_path, "r", encoding=ENCODING) as file:
        return json.loads(file.read())


def read_text_from_file(text_file_path: str) -> str:
    with open(text_file_path, "r", encoding=ENCODING) as file:
        return file.read()


def write_text_to_file(text: str, file_path: str):
    with open(file_path, "w", encoding=ENCODING) as file:
        file.write(text)


def read_object_from_yaml_file(yaml_file_path: str):
    with open(yaml_file_path, "r", encoding=ENCODING) as file:
        return yaml.load(file.read(), Loader=yaml.FullLoader)


class NoAliasDumper(yaml.Dumper):
    def ignore_aliases(self, data):
        return True


def write_object_to_yaml_file(obj: Union[dict, list], yaml_file_path: str):
    with open(yaml_file_path, "w", encoding=ENCODING) as file:
        yaml.dump(obj, file, Dumper=NoAliasDumper)


def read_lines_from_txt_file(txt_file_path: str) -> Iterator[str]:
    with open(txt_file_path, "r", encoding=ENCODING) as file:
        for line in file:
            yield line.rstrip()


def count_lines_from_txt_file(txt_file_path: str):
    with open(txt_file_path) as f:
        for i, l in enumerate(f):
            pass
    return i + 1


def load_excel(excel_path, from_row: int = 1) -> DataFrame:
    arias_scoring = pd.ExcelFile(excel_path)
    return arias_scoring.parse(skiprows=from_row - 1)


def combine_lists(
    first_element: str, second_list: str, third_list: List[str]
) -> List[str]:
    final_list = []

    for s in third_list:
        if second_list != "":  # in 1-element combinations
            final_list.append(first_element + "," + second_list + "," + s)
        else:
            final_list.append(first_element + "," + s)

    return final_list


def read_dicts_from_csv(file_path: str) -> List[dict]:
    with open(file_path, "r", encoding=ENCODING) as file:
        return [obj for obj in csv.DictReader(file)]


def write_dicts_to_csv(dicts: List[dict], file_path: str):
    with open(file_path, "w", encoding=ENCODING) as file:
        writer = csv.DictWriter(
            file,
            fieldnames=list(dicts[0].keys()),
            delimiter=CSV_DELIMITER,
        )
        writer.writeheader()
        for i, obj in enumerate(dicts):
            writer.writerow(obj)


def get_color(levelname: str) -> str:
    return COLOR_SEQ % (30 + COLORS[levelname])


def colorize(text: str, levelname: str):
    return get_color(levelname) + text + RESET_SEQ


def extract_digits(text: str) -> str:
    return "".join([char for char in text if char.isdigit()])



import builtins
import pickle
import weakref
from pathlib import PurePath, Path
from typing import Any, List, Optional, Tuple

import music21 as m21
import pandas as pd

from musif.cache.cache import MethodCache, ObjectReference, SmartModuleCache


class FileCacheIntoRAM:
    """
    This class simply stores a dictionary of key-value. In `musif`, it is used to cache
    the objects (values) coming from the parsing of files (whose names are the keys). It
    is never written to disk and only kept into RAM.
    """

    def __init__(self, capacity: int):
        self._capacity = capacity
        self._items = []

    def put(self, key: str, value: Any):
        if self.full:
            del self._items[0]
        self._items.append({"key": key, "value": value})

    def get(self, key: str) -> Optional[Any]:
        for item in self._items:
            if item["key"] == key:
                return item["value"]
        return None

    def clear(self) -> None:
        self._items.clear()

    @property
    def full(self) -> bool:
        return len(self._items) == self._capacity


def iscache(obj1):
    """
    Check if `obj1` is a `SmartModuleCache` or a `MethodCache` object
    """
    if type(obj1) in (MethodCache, SmartModuleCache):
        return True
    return False


def isinstance(obj1, cls):
    """
    Check if obj1 is instance of `cls`. This function grants that `SmartModuleCache`
    objects are checked against their reference objects.
    """
    if type(obj1) is SmartModuleCache:
        t = obj1.cache["_type"]
        return builtins.issubclass(t, cls)
    else:
        return builtins.isinstance(obj1, cls)


def hasattr(obj1, attr):
    """
    Check if `obj1` has `attr`. This function grants that `SmartModuleCache` objects are
    checked against their cache or reference objects.
    """
    if type(obj1) is SmartModuleCache:
        ref = obj1.cache["_reference"].reference
        if ref is not None:
            return builtins.hasattr(ref, attr)
        else:
            return attr in obj1.cache
    else:
        return builtins.hasattr(obj1, attr)


def wrap_module_objects(
    obj: Any,
    target_addresses: List[str] = ["music21"],
    resurrect_reference: Optional[Tuple] = None,
    parent: Optional[ObjectReference] = None,
    name: Tuple[str] = ("",),
    args: Tuple[Optional[Tuple]] = (None,),
):
    """
    Returns the object wrapped with `SmartModuleCache` class if it was defined
    in one of the `target_addresses`

    If `obj` is a list or a tuple, e new list/tuple this function works
    recursively on their objects.

    If the object is an instance of `weakref.ReferenceType`
    this function converts the object to a regular
    reference and then applies the wrapping.
    """
    if isinstance(obj, weakref.ReferenceType):
        __import__('ipdb').set_trace()
        return obj()

    __module = obj.__class__.__module__
    for module in target_addresses:
        if __module.startswith(module):
            return SmartModuleCache(
                reference=obj,
                target_addresses=target_addresses,
                resurrect_reference=resurrect_reference,
                parent=parent,
                name=name,
                args=args,
            )

    if isinstance(obj, (list, tuple)):
        ret = [
            wrap_module_objects(
                v,
                target_addresses,
                resurrect_reference,
                parent,
                (*name, "__getitem__"),
                (*args, (i,)),
            )
            for i, v in enumerate(obj)
        ]
        if isinstance(obj, list):
            return ret
        elif isinstance(obj, tuple):
            return tuple(ret)
    return obj


def store_score_df(score, fname):
    """
    Stores `score` into `fname` (a file-like object, a string or a Path object)
    using dataframes and returns the object saved.

    The returned object is a dictionary with keys the name of the parts and values
    dataframes with the following columns:

    * "Type": A string identifying the type of object. Possible values: ``"Note"``,
        ``"Rest"``, ``"Measure"``,  ``"Time Signature"``
    * "Name": A string with the name of the note in Common Western Notation or with
        the time signature string for time signatures; for measures and rests, the value
        ``"-"`` is used.
    * "Value": The midi pitch for notes, -1 for others
    * "Measure Onset": The beat position of the object in reference to the beginning
        of the measure, -1 for measures
    * "Part Onset": The onset position of the object in reference to the beginning
        of the part
    * "Duration": The duration of the object, -1 for time signatures
    * "Tie": If a tie is applied to the note, its type is there (one of ``"start"``,
        ``"continue"``, ``"stop"``), otherwise ``"-"`` is used
    """

    def append_note(alist, note, offset, type, pitch=None, name=None):
        onset = offset + note.offset
        dur = note.duration.quarterLength
        m_onset = note.offset
        if pitch is None:
            pitch = note.pitch.midi
        if name is None:
            name = note.nameWithOctave
        if note.tie is not None:
            tie = note.tie.type
        else:
            tie = "-"
        alist.append((type, name, pitch, m_onset, onset, dur, tie))

    score_dict = {}
    for part in score.parts:
        data_part = []
        for measure in part.getElementsByClass(m21.stream.base.Measure):
            offset = measure.offset
            data_part.append(("Measure", "-", -1, -1, offset,
                              measure.duration.quarterLength, "-"))
            ts = measure.timeSignature
            if ts is not None:
                data_part.append(
                    (
                        "Time Signature",
                        ts.ratioString,
                        -1,
                        ts.offset,
                        ts.offset + offset,
                        -1,
                        "-",
                    )
                )
            for note in measure.flat.notesAndRests:
                if isinstance(note, m21.note.Note):
                    append_note(data_part, note, offset, "Note")
                elif isinstance(note, m21.note.Rest):
                    append_note(data_part, note, offset, "Rest", name="-", pitch=-1)
                elif isinstance(note, m21.chord.Chord):
                    for note in note.notes:
                        append_note(data_part, note, offset, "Note")
                else:
                    raise Exception(
                        f"Cannot handle this type of note: {note.cache['_type']}"
                    )
        data_part = pd.DataFrame(
            data_part,
            columns=(
                "Type",
                "Name",
                "Value",
                "Measure Onset",
                "Part Onset",
                "Duration",
                "Tie",
            ),
        )
        score_dict[part.partName] = data_part
    if isinstance(fname, (PurePath, str)):
        fname = Path(fname)
        fname.parent.mkdir(exist_ok=True)
        fname = open(fname, "wb")
    pickle.dump(score_dict, fname)
    return score_dict


def load_score_df(fname):
    """
    Loads a score object saved with `store_score_df` from a string or Path pointing to
    the file or from a file-like object. Return a dictionary of dataframes
    """
    if isinstance(fname, (PurePath, str)):
        fname = Path(fname)
        fname = open(fname, "rb")
    return pickle.load(fname)



from musif.extract.features.prefix import get_part_prefix

PRESENCE = "Presence_of"
"""Constant to create columns regarding presenc of an instrument or not."""

voices_list = ["sop", "ten", "alt", "bar", "bbar", "bass"]
"""List of prefixes for singers used to find voices columns"""

voices_list_prefixes = [get_part_prefix(i) for i in voices_list]
"""List of prefixes `Part_` + each element of voices_list"""



import re
import xml.etree.ElementTree as ET
from enum import Enum
from typing import Optional

import pandas as pd


class TempoGroup2(Enum):
    """
    Contains constants refering to grouped tempo on the second level that will be used by the tempo module.
    """

    NA = "NA"
    SLOW = "Slow"
    MODERATE = "Moderate"
    FAST = "Fast"


def get_time_signature_type(time_signature: str) -> str:
    """
    Given a time signature, it returns  the correspondent type.

    Parameters
    ----------
        time_signature: str
        Time signature in the form of string separated by '/' to obtain the time signature type.
    """
    if time_signature in [
        "1/2",
        "1/4",
        "1/8",
        "1/16",
        "2/2",
        "2/4",
        "2/8",
        "2/16",
        "4/4",
        "C",
        "4/2",
        "4/8",
        "4/16",
        "8/2",
        "8/4",
        "8/8",
        "8/16",
    ]:
        return "simple duple"

    elif time_signature in ["6/8", "3/8", "12/2", "12/4", "12/8", "12/16"]:
        return "compound duple"

    elif time_signature in ["3/2", "3/4", "3/16", "6/2", "6/4", "6/16"]:
        return "simple triple"

    elif time_signature in ["9/2", "9/4", "9/8", "9/16"]:
        return "compound triple"

    else:
        return "other"


def get_tempo_grouped_1(tempo: str) -> str:
    """
    Returns a 1st level of grouping for the tempo markings, removing secondary labels
    and diminutive endings.

    Parameters
    ----------
        tempo: str
        Tempo string in italian to be extract the group from.

    """
    tempo = re.sub("\\W+", " ", tempo)  # removes eventual special characters
    replacements = [
        (w, "")
        for w in [
            "molto",
            "poco",
            "un poco",
            "tanto",
            "un tanto",
            "assai",
            "meno",
            "più",
            "piuttosto",
        ]
    ]
    if not tempo:
        return "NA"

    for r in replacements:
        tempo = tempo.replace(*r)
    tempo = (
        tempo.replace("Con brio", "brio")
        .replace("con brio", "brio")
        .replace("Con spirito", "spiritoso")
        .replace("con spirito", "spiritoso")
    )
    base_important_words = [
        "adagio",
        "allegro",
        "andante",
        "andantino",
        "largo",
        "lento",
        "presto",
        "vivace",
        "minueto",
    ]
    # if the tempo marking ends in -ietto and or -issimo, it retuns the marking without that ending
    important_words = (
        base_important_words
        + [w[0:-1] + "etto" for w in base_important_words]
        + [w[0:-1] + "ietto" for w in base_important_words]
        + [w[0:-1] + "issimo" for w in base_important_words]
        + [w[0:-1] + "ssimo" for w in base_important_words]
        + [w[0:-1] + "hetto" for w in base_important_words]
    )

    last_words = ""
    for word in tempo.split(" "):
        if word.lower() in important_words and (
            "ma non" not in last_words
            or " ma " not in last_words
            or "Tempo di" in last_words
        ):
            return word.capitalize()
        last_words += " " + word

    ################ SECOND IMPORTANT WORDS ################
    relevant_words = [
        "amoroso",
        "affettuoso",
        "agitato",
        "arioso",
        "cantabile",
        "comodo",
        "brio",
        "spiritoso",
        "espressivo",
        "fiero",
        "giusto",
        "grave",
        "grazioso",
        "gustoso",
        "maestoso",
        "moderato",
        "risoluto",
        "sostenuto",
        "spiritoso",
        "tempo",
    ]
    words_contained = []
    for word in tempo.split(" "):
        if word.lower() in relevant_words:
            words_contained.append(word)
    ## FINAL DECISION (Two exceptions: tempo and con brio) ##
    if len(words_contained) == 1:
        if words_contained[0].lower() not in ["tempo", "brio"]:
            return words_contained[0].capitalize()
        elif words_contained[0].lower() == "tempo":
            return "A tempo"
        else:
            return "Con brio"
    elif len(words_contained) > 1:
        for w in words_contained:
            if w[0].isupper():
                if w.lower() != "tempo":
                    return w  # returns the capitalized term
                elif w.lower() == "tempo":
                    return "A tempo"
                else:
                    return "Con brio"
        if words_contained[0] != "tempo":
            return words_contained[0].capitalize()
        elif words_contained[0] == "tempo":
            return "A tempo"  # or the first one
        else:
            return "Con brio"

    return "NA"


def get_tempo_grouped_2(tempo_grouped_1: str) -> str:
    """
    Returns a 2nd level of grouping for the tempo markings.

    Parameters
    ----------
        tempo_grouped_1: str
        Tempo string got from get_tempo_grouped_1
    """
    if tempo_grouped_1 is None or tempo_grouped_1.lower() == TempoGroup2.NA.value:
        return TempoGroup2.NA.value
    possible_terminations = ["ino", "etto", "ietto", "ssimo", "issimo", "hetto"]
    slow_basis = [
        "Adagio",
        "Affettuoso",
        "Grave",
        "Sostenuto",
        "Largo",
        "Lento",
        "Sostenuto",
    ]
    slow = slow_basis + [w[:-1] + t for w in slow_basis for t in possible_terminations]
    moderate_basis = [
        "Andante",
        "Arioso",
        "Comodo",
        "Cantabile",
        "Comodo",
        "Espressivo",
        "Grazioso",
        "Gustoso",
        "Maestoso",
        "Minueto",
        "Moderato",
        "Marcía",
        "Amoroso",
    ]
    moderate = moderate_basis + [
        w[:-1] + t for w in moderate_basis for t in possible_terminations
    ]
    fast_basis = [
        "Agitato",
        "Allegro",
        "Con brio",
        "Spiritoso",
        "Fiero",
        "Presto",
        "Risoluto",
        "Vivace",
    ]
    fast = fast_basis + [w[:-1] + t for w in fast_basis for t in possible_terminations]

    if tempo_grouped_1 in ["A tempo", "Giusto"]:
        return TempoGroup2.NA.value
    elif tempo_grouped_1 in slow:
        return TempoGroup2.SLOW.value
    elif tempo_grouped_1 in moderate:
        return TempoGroup2.MODERATE.value
    elif tempo_grouped_1 in fast:
        return TempoGroup2.FAST.value


def get_number_of_beats(time_signature: str) -> int:
    """
    Returns the number of beats corresponding to a time signature.

    By default, this returns the numerator, with the exceptions when the numerator is
    6, 9, or 12 and when the time signature is C or 3/8. If it is empty, it returns 1.
    If it is `"NA"`, it returns `pd.NA`.

    Parameters
    ----------
        time_signature: str
            Time signature in the form of string separated by '/' to obtain the time
            signature type.

    """
    # we only take the first time signature in case we receive more than one
    time_signature = time_signature.split(",")[0]
    if time_signature == "C":
        return 4
    elif time_signature == "":
        return 1
    elif time_signature == "NA":
        return pd.NA
    num, den = time_signature.split("/")
    if num == "3" and den == "8":
        return 1
    elif num == "6":
        return 2
    elif num == "9":
        return 3
    elif num == "12":
        return 4
    else:
        return int(num)
        # raise ValueError(f"The {time_signature} is not a known time signature")


def extract_numeric_tempo(file_path: str) -> Optional[int]:
    """
    Finds the numeric tempo in a musixml file by looking at the tempo marking in
    the xml code.

    Parameters
    ----------

        file_path: str
        Path to xml file to get the tempo from.

    """
    try:
        tree = ET.parse(file_path)
    except ET.ParseError as e:
        return "NA"

    root = tree.getroot()
    try:
        tempo = int(
            root.find("part")
            .find("measure")
            .find("direction")
            .find("sound")
            .get("tempo")
        )
    except:
        tempo = "NA"
    return tempo



import pickle
from typing import Dict

# import nltk
# from googletrans import Translator
# nltk.download('wordnet', quiet=True)
# from nltk.corpus import wordnet as wn
from musif.common._utils import read_object_from_json_file


def translate_word(original_word: str, translations: dict) -> str:
    parenthesis = original_word.find("(")
    word = original_word[:parenthesis] if "(" in original_word else original_word
    word = word.lower().strip()

    for key, value in translations.items():
        if word in value:
            return key.capitalize()

    return original_word

    # def translate_word(word: str, language: str = None, translations_cache: Dict[str, str] = None):
    #     t = Translator()
    translator_x_eng = ""
    # Remove instrument specifications (Re), (Do), ...
    parenthesis = word.find("(")
    word = word[:parenthesis] if "(" in word else word
    # translations_cache = translations_cache or {}
    # cached_translation = translations_cache.get(word.lower().strip())
    # if cached_translation:
    #     return cached_translation
    for i in range(3):  # to support errors in the
        # try:
        # translator_x_eng = translations_cache.get(word.lower().strip(), t.translate(word, src='auto' if language ==
        # translator_x_eng_text = translator_x_eng.text
        break


#         except:
#             translator_x_eng_text = word
# from the list of possible translations -> take nouns only
#     main_possibility = translator_x_eng_text.replace('-', ' ').lower()
#     # singularize(main_possibility)
#     if len(wn.synsets(main_possibility)) > 0 and wn.synsets(main_possibility)[0].pos() != 'n':  # if not noun: look for a noun translation
#         try:
#             # if there is more than one possible translation:
#             for translation in translator_x_eng.extra_data['all-translations']:
#                 type_word = translation[0]
#                 if type_word == 'noun':
#                     main_possibility = translation[1][0].replace('-', ' ').lower()
#         except:
#             pass
#     return main_possibility
